
(require-extension (srfi 10000))

(common-module (common process)
  ((export process-run process-run-and-exit
           process-output->string process-output->string-list
           with-input-from-process with-output-to-process
           with-process-io))

(define (shell-parse str)
  (let ((len (string-length str)))
    (define (outside i res)
      (if (>= i len)
        (reverse res)
        (let ((c (string-ref str i)))
          (case c
            ((#\space #\tab #\newline) (outside (+ i 1) res))
            ((#\" #\') (quoted (+ i 1) c '() res))
            ((#\\) (inside (+ i 2) (list (string-ref str (+ i 1))) res))
            (else (inside (+ i 1) (list c) res))))))
    (define (inside i ls res)
      (if (>= i len)
        (reverse (cons (list->string (reverse ls)) res))
        (let ((c (string-ref str i)))
          (case c
            ((#\space #\tab #\newline)
             (outside (+ i 1) (cons (list->string (reverse ls)) res)))
            ((#\" #\')
             (quoted (+ i 1) c ls res))
            ((#\\)
             (inside (+ i 2) (cons (string-ref str (+ i 1)) ls) res))
            (else
             (inside (+ i 1) (cons c ls) res))))))
    (define (quoted i terminal ls res)
      (if (>= i len)
        (reverse (cons (list->string (reverse ls)) res))
        (let ((c (string-ref str i)))
          (cond
            ((eqv? c terminal)
             (inside (+ i 1) ls res))
            ((eqv? c #\\)
             (quoted (+ i 2) terminal (cons (string-ref str (+ i 1)) ls) res))
            (else
             (quoted (+ i 1) terminal (cons c ls) res))))))
    (outside 0 '())))

(define (expand-shell-command str)
  (define (string-split-char str . o)
    (let ((sep (if (pair? o) (car o) #\space))
          (len (string-length str)))
      (let lp ((i 0) (j 0) (res '()))
        (define (collect) (cons (substring str j i) res))
        (cond ((= i len) (reverse (collect)))
              ((eqv? sep (string-ref str i)) (lp (+ i 1) (+ i 1) (collect)))
              (else (lp (+ i 1) j res))))))
  (let* ((command (shell-parse (display-to-string str)))
         (bin (car command)))
    (if (or (equal? bin "") (eqv? #\/ (string-ref bin 0)))
      command
      (let lp ((ls (string-split-char (getenv "PATH") #\:)))
        (if (null? ls)
          command
          (let ((path (string-append (car ls) "/" bin)))
            (if (file-exists? path)
              (cons path (cdr command))
              (lp (cdr ls)))))))))

(define (shell-escape-char? c)
  (not (or (char-alphabetic? c)
           (char-numeric? c)
           (memv c '(#\_ #\- #\+ #\: #\, #\= #\/ #\.)))))

(define (shell-escape-backslash str)
  (with-output-to-string
    ""
    (lambda ()
      (with-input-from-string str
        (lambda ()
          (let lp ()
            (let ((c (read-char)))
              (if (not (eof-object? c))
                (begin
                  (if (shell-escape-char? c)
                    (display #\\))
                  (display c)
                  (lp))))))))))

(define (shell-escape str)
  (let lp ((i (- (string-length str) 1)) (quote? #f))
    (if (negative? i)
      (if quote?
        (string-append "\"" str "\"")
        str)
      (let ((c (string-ref str i)))
        (cond
          ((or (eqv? c #\") (eqv? c #\\))
           (shell-escape-backslash str))
          ((shell-escape-char? c)
           (lp (- i 1) #t))
          (else
           (lp (- i 1) quote?)))))))

(define (display-to-string x)
  (if (string? x)
    x
    (with-output-to-string "" (lambda () (display x)))))

(define (string-intersperse ls . o)
  (if (null? ls)
    ""
    (let ((str (if (pair? o) (car o) " ")))
      (let lp ((ls (cdr ls)) (res (list (car ls))))
        (if (null? ls)
          (apply string-append (reverse res))
          (lp (cdr ls) (cons (car ls) (cons str res))))))))

(define (make-process-string command ls)
  (define (escape x)
    (shell-escape (if (string? x) x (object->string x))))
  (string-intersperse (cons (display-to-string command)
                            (map escape ls))))

(define (process-run command . args)
  (zero? (shell-command (make-process-string command args))))

(define (process-run-and-exit command . args)
  (exit (shell-command (make-process-string command args))))

(define (%open-process-port x)
  (let ((ls (if (pair? x)
              (append (expand-shell-command (car x)) (cdr x))
              (expand-shell-command x))))
    (if (pair? (cdr ls))
      (open-process path: (car ls) arguments: (cdr ls))
      (open-process (car ls)))))

(define (with-input-from-process x thunk)
  (let* ((p (%open-process-port x))
         (res (with-input-from-port p thunk)))
    (close-port p)
    res))

(define (with-output-to-process x thunk)
  (let* ((p (%open-process-port x))
         (res (with-output-to-port p thunk)))
    (close-port p)
    res))

(define (with-process-io x thunk)
  (let* ((p (%open-process-port x))
         (res (with-input-from-port p
                (lambda ()
                  (with-output-to-port p
                    thunk)))))
    (close-port p)
    res))

(define (process-output->string-list x . args)
  (with-input-from-process x
    (lambda ()
      (let lp ((res '()))
        (let ((line (read-line)))
          (if (eof-object? line)
            (reverse res)
            (lp (cons line res))))))))

(define (process-output->string . o)
  (string-intersperse (apply process-output->string-list o) "\n"))

)
