
(define-module (common pathname))

(export path-absolute? path-strip-directory path-directory
        path-extension path-strip-extension path-replace-extension
        path-expand path-normalize path-relative path-build
        current-directory)

(define current-directory getcwd)
(define set-current-directory! chdir)

(define (display-to-string x)
  (if (string? x) x (with-output-to-string (lambda () (display x)))))

(define (string-scan c str . o)
  (let ((pred (if (char? c) (lambda (x) (eqv? x c)) c)))
    (let ((limit (string-length str)))
      (let lp ((i (or (and (pair? o) (car o)) 0)))
        (cond ((>= i limit) #f)
              ((pred (string-ref str i)) i)
              (else (lp (+ i 1))))))))

(define (string-scan-right c str . o)
  (let ((pred (if (char? c) (lambda (x) (eqv? x c)) c)))
    (let lp ((i (or (and (pair? o) (car o)) (- (string-length str) 1))))
      (cond ((negative? i) #f)
            ((pred (string-ref str i)) i)
            (else (lp (- i 1)))))))

(define (path-strip-directory path)
  (if (string=? path "")
    path
    (let ((len (string-length path)))
      (if (eqv? #\/ (string-ref path (- len 1)))
        ""
        (let ((slash (string-scan-right #\/ path)))
          (if (not slash)
            path
            (substring path (+ slash 1) len)))))))

(define (path-directory path)
  (if (string=? path "")
    "."
    (let ((end (string-scan-right (lambda (c) (not (eqv? c #\/))) path)))
      (if (not end)
        "/"
        (let ((start (string-scan-right #\/ path (- end 1))))
          (if (not start)
            "."
            (let ((start
                   (string-scan-right
                    (lambda (c) (not (eqv? c #\/))) path start)))
              (if (not start) "/" (substring path 0 (+ start 1))))))))))

(define (path-extension-pos path)
  (string-scan-right (lambda (c) (or (eqv? c #\.) (eqv? c #\/))) path))

(define (path-extension path)
  (let ((i (path-extension-pos path)))
    (and i
         (not (eqv? (string-ref path i) #\/))
         (let ((start (+ i 1))
               (end (string-length path)))
           (and (< start end) (substring path start end))))))

(define (path-strip-extension path)
  (let ((i (path-extension-pos path)))
    (if (and i
             (< (+ i 1) (string-length path))
             (not (eqv? (string-ref path i) #\/)))
      (substring path 0 i)
      path)))

(define (path-replace-extension path ext)
  (string-append (path-strip-extension path) "." ext))

(define (path-expand path . o)
  (define (get-user-home name)
    (if (string=? name "") (getenv "HOME") (string-append "/home/" name)))
  (define (here) (if (pair? o) (car o) (current-directory)))
  (if (string=? path "")
    (here)
    (case (string-ref path 0)
      ((#\~)
       (let* ((off (string-scan #\/ path 1))
              (name (if off (substring path 1 off) (substring path 1))))
         (string-append
          (get-user-home name) (if off (substring path off) ""))))
      ((#\/) path)
      (else (string-append (here) "/" path)))))

(define (path-absolute? path)
  (and (not (string=? "" path)) (eqv? #\/ (string-ref path 0))))

;; Currently just returns #f if path is not within the enclosing path,
;; may want to return a relative path via "..".

(define (path-relative path . o)
  (let* ((path (path-normalize (path-expand path)))
         (p-len (string-length path))
         (dir (path-normalize (if (or (null? o) (equal? (car o) "."))
                                (current-directory)
                                (path-expand (car o)))))
         (d-len (string-length dir)))
    (and (> p-len d-len)
         (string=? dir (substring path 0 d-len))
         (cond
           ((eqv? #\/ (string-ref path d-len))
            (substring path (+ d-len 1) p-len))
           ((eqv? #\/ (string-ref path (- d-len 1)))
            (substring path d-len p-len))
           (else #f)))))

;; Mutation-free and guarantees:
;;   (string=? s (path-normalize s))  <=>  (eq? s (path-normalize s))
;; i.e. fast and simple for already normalized paths.

(define (path-normalize path)
  (let* ((len (string-length path)) (len-1 (- len 1)))
    (define (collect i j res)
      (if (>= i j) res (cons (substring path i j) res)))
    (define (finish i res)
      (if (zero? i)
        path
        (apply string-append (reverse (collect i len res)))))
    (define (backup j s res)
      (let ((pos (+ j 3)))
        (cond
          ((or (string=? s "/..") (string=? s ".."))
           (boundary pos pos (cons "/.." (cons s res))))
          ((or (string=? s "") (string=? s ".") (string=? s "/"))
           (if (pair? res)
             (backup j (car res) (cdr res))
             (boundary pos pos (if (string=? s "/") '("/") '()))))
          (else
           (let ((d (path-directory s)) (pos (+ j 3)))
             (cond
               ((string=? d "/")
                (boundary pos pos (if (null? res) '("/") res)))
               ((string=? d ".")
                (if (pair? res)
                  (boundary pos pos res)
                  (boundary pos pos (list "/.." s))))
               (else (boundary pos pos (cons "/" (cons d res))))))))))
    (define (boundary i j res)
      (if (>= j len-1)
        (finish i res)
        (case (string-ref path j)
          ((#\.)
           (case (string-ref path (+ j 1))
             ((#\.)
              (if (or (>= j (- len 2)) (eqv? #\/ (string-ref path (+ j 2))))
                (if (>= i (- j 1))
                  (if (null? res)
                    (backup j "" '())
                    (backup j (car res) (cdr res)))
                  (backup j (substring path i j) res))
                (inside i (+ j 2) res)))
             ((#\/)
              (if (= i j)
                (boundary (+ j 2) (+ j 2) res)
                (let ((s (substring path i j)))
                  (boundary (+ j 2) (+ j 2) (cons s res)))))
             (else (inside i (+ j 1) res))))
          ((#\/) (boundary (+ j 1) (+ j 1) (collect i j res)))
          (else (inside i (+ j 1) res)))))
    (define (inside i j res)
      (if (>= j len)
        (finish i res)
        (if (eqv? #\/ (string-ref path j))
          (boundary i (+ j 1) res)
          (inside i (+ j 1) res))))
    (if (zero? len)
      path
      ((if (eqv? #\/ (string-ref path 0)) boundary inside) 0 1 '()))))

(define (path-build . args)
  (define (trim-trailing-slash s)
    (let ((i (string-scan-right (lambda (c) (not (eqv? c #\/))) s)))
      (if i (substring s 0 (+ i 1)) "")))
  (if (null? args)
    ""
    (let ((start (trim-trailing-slash (display-to-string (car args)))))
      (let lp ((ls (cdr args))
               (res (if (string=? "" start) '() (list start))))
        (cond
          ((null? ls)
           (apply string-append (reverse res)))
          ((pair? (car ls))
           (lp (append (car ls) (cdr ls)) res))
          (else
           (let ((x (trim-trailing-slash (display-to-string (car ls)))))
             (lp (cdr ls) (if (string=? x "")
                            res
                            (cons x (cons "/" res)))))))))))
