
(define-module (srfi srfi-10000)
  #:export (common-module import-extension import-syntax import-lazy
                          let-optionals* let-keywords*))

(define-macro (common-module name clauses . body)
  (generate-common-module
   (if (null? name) (list 'anonymous (gensym)) name) clauses body))

(define-macro (import-extension . args)
  `(use-modules ,@(expand-module-normal-list args)))

(define-macro (import-syntax . args)
  `(import-extension ,@args))

(define-macro (import-lazy mod . args)
  `(import-extension ,mod))

(define (x->string x)
  (cond ((number? x) (number->string x))
        ((symbol? x) (symbol->string x))
        ((char? x) (string x))
        (else x)))

(define (x->module-name x)
  (if (pair? x)
    x
    (map string->symbol (string-split (x->string x) #\/))))

(define srfi-aliases
  '((cond-expand . 0)
    (list-lib . 1)
    (and-let* . 2)
    (numeric-vectors . 4)
    (signature-let . 5)
    (string-ports . 6)
    (configuration-language . 7)
    (receive . 8)
    (define-record-type . 9)
    (external-form-syntax . 10)
    (multiple-value-bind . 11)
    (string-lib . 13)
    (char-set-lib . 14)
    (case-lambda . 16)
    (generalized-set! . 17)
    (multithreading . 18)
    (time-lib . 19)
    (real-time-multithreading . 21)
    (unix-scheme-scripts . 22)
    (error . 23)
    (array-lib . 25)
    (cut . 26)
    (random-numbers . 27)
    (basic-format . 28)
    (localization . 29)
    (multi-line-comment-syntax . 30)
    (rec . 31)
    (sort-lib . 32)
    (bitwise-lib . 33)
    (exceptions . 34)
    (conditions . 35)
    (i/o-conditions . 36)
    (args-fold . 37)
    (shared-structure-read/write . 38)
    (parameters . 39)
    (streams-lib . 40)
    (comprehensions . 42)
    (vector-lib . 43)
    (comprehensions . 44)
    (lazy-eval . 45)
    (syntax-rules-choose-ellipse . 46)
    (homogeneous-array-lib . 47)
    (intermediate-format . 48)
    (require-extension . 55)
    (advanced-records . 57)
    (array-notation . 58)
    (vicinity . 59)
    (bitwise-operations . 60)
    (sexpr-comments . 62)
    (homogeneous-and-heterogeneous-arrays . 63)
    (testing . 64)
    ))

(define (any1 pred ls)
  (let lp ((ls ls)) (and (pair? ls) (or (pred (car ls)) (lp (cdr ls))))))

(define (append-map1 f ls)
  (let lp ((ls ls) (res '()))
    (if (null? ls)
      (reverse res)
      (lp (cdr ls) (append (reverse (f (car ls))) res)))))

(define (expand-module-names spec)
  (cond
    ((null? spec) (list spec))
    ((not (pair? spec)) (list (list spec)))
    (else
     (append-map1
      (lambda (head)
        (map (lambda (tail) (append head tail))
             (expand-module-names (cdr spec))))
      (if (pair? (car spec))
        (append-map1 expand-module-names (car spec))
        (list (list (car spec))))))))

(define (expand-module-name-list ls)
  (append-map1 expand-module-names ls))

(define (expand-module-normal-list ls)
  (map module-name-normalize (expand-module-name-list ls)))

(define (module-name-normalize ls)
  (if (and (eq? 'srfi (car ls)) (pair? (cdr ls)) (null? (cddr ls)))
    (list 'srfi
          (string->symbol
           (string-append
            "srfi-"
            (x->string (cond ((assq (cadr ls) srfi-aliases) => cdr)
                             (else (cadr ls)))))))
    ls))

(define (generate-common-module name clauses body)
  (define (expand-cond clauses)
    (define (every1 pred ls)
      (not (any1 (lambda (x) (not (pred x))) ls)))
    (define (true-clause? x)
      (case x
        ((guile else) #t)
        (else
         (and (pair? x)
              (case (car x)
                ((and) (every1 true-clause? (cdr x)))
                ((or)  (any1 true-clause? (cdr x)))
                ((not) (not (true-clause? (cadr x))))
                (else #f))))))
    (cond
      ((null? clauses) '())
      ((true-clause? (caar clauses)) (cdar clauses))
      (else (expand-cond (cdr clauses)))))
  (define (expand-imports ls)
    (let lp ((ls ls) (res '()))
      (cond
        ((null? ls)
         (map module-name-normalize (reverse res)))
        ((memq (caar ls) '(import import-only import-syntax import-extension))
         (lp (cdr ls) (append (expand-module-name-list (cdar ls)) res)))
        ((eq? (caar ls) 'import-lazy)
         (lp (cdr ls)
             (cons (car (expand-module-names (cadar ls)))
                   res)))
        ((eq? (caar ls) 'import-rename)
         (let* ((mod-name (module-name-normalize (cadar ls)))
                (local-name (caddar ls))
                (exported-name (car (cdddar ls))))
           (lp (cdr ls)
               (cons `(,mod-name #:select ((,exported-name . ,local-name)))
                     res))))
        ((eq? (caar ls) 'import-prefix)
         (let* ((mod-name (module-name-normalize (cadar ls)))
                (prefix (caddar ls)))
           (lp (cdr ls)
               (cons `(,mod-name #:renamer (symbol-prefix-proc ',prefix))
                     res))))
        (else
         (lp (cdr ls) res)))))
  (let lp ((ls clauses)
             (imps '())
             (exps '())
             (main #f)
             (body body))
    (if (pair? ls)
      (case (if (pair? (car ls)) (caar ls) 'dummy)
        ((entry-point)
         (lp (cdr ls) imps exps (cadar ls) body))
        ((export export-syntax)
         (lp (cdr ls) imps (append (cdar ls) exps) main body))
        ((import import-only import-extension import-syntax import-lazy
                 import-prefix import-rename)
         (lp (cdr ls) (cons (car ls) imps) exps main body))
        ((inherit)
         (let* ((mods (expand-module-normal-list (cdar ls)))
                (exports
                 (delete
                  '%module-public-interface
                  (append-map1
                   (lambda (m)
                     (module-map (lambda (x . o) x) (resolve-module m)))
                   mods))))
           (lp (cdr ls) (cons `(import-extension ,@(cdar ls)) imps)
               (append exps exports)
               main
               (append
                (append-map1
                 (lambda (x)
                   (map
                    (lambda (s)
                      `(define ,s (module-ref (resolve-module ',x) ',s)))
                    (delete '%module-public-interface
                            (module-map (lambda (x . o) x) (resolve-module x)))))
                 mods)
                body))))
        ((cond)
         (lp (append (expand-cond (cdar ls)) (cdr ls)) imps exps main body))
        ((keywords)
         (lp (cdr ls) imps (append (cdar ls) exps) main
             (append (map (lambda (s) `(define ,s ',s)) (cdar ls)) body)))
        ((prelude postlude prelude-include postlude-include)
         (lp (cdr ls) imps exps main
             ((case (caar ls)
                ((prelude prelude-include) append)
                (else (lambda (a b) (append b a))))
              (case (caar ls)
                ((prelude postlude) (cdar ls))
                (else `((load ,(cadar ls)))))
              body)))
        (else (lp (cdr ls) imps exps main body)))
      (let ()
        `(begin
           (define-module ,(x->module-name name))
           (export ,@exps)
           (use-modules (srfi srfi-10000)
                        (ice-9 syncase)
                        ,@(expand-imports imps))
           ,@body
           ,@(if main `((,main (command-line))) '()))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-macro (let-optionals* ls vars . body)
  (if (null? vars)
    `(let () ,@body)
    (let ((tmp (gensym)))
      `(let* ((,tmp ,ls)
              (,(caar vars)
               (if (pair? ,tmp) (car ,tmp) ,(cadar vars))))
         (let-optionals* (if (pair? ,tmp) (cdr ,tmp) '()) ,(cdr vars)
           ,@body)))))

(define (proper-list? x)
  (or (null? x) (and (pair? x) (proper-list? (cdr x)))))

(define (with-split-improper-list ls proc)
  (let loop ((l ls) (res '()))
    (cond ((pair? l) (loop (cdr l) (cons (car l) res)))
          (else (proc (reverse! res) l)))))

(define (with-split-at-sym sym ls proc)
  (let lp ((ls ls) (head '()))
    (if (or (null? ls) (eq? sym (car ls)))
      (proc (reverse head) ls)
      (lp (cdr ls) (cons (car ls) head)))))

(define (x->keyword x)
  (let ((str (x->string x)))
    (string->symbol (string-append str ":"))))

(define (extract-spec x l)
  (cond
    ((symbol? x)
     (extract-spec (list x) l))
    ((and (list? x) (<= 1 (length x) 4) (symbol? (car x)))
     (let ((var (car x)))
       (with-split-at-sym '=> (cdr x)
         (lambda (info act)
           (let ((keys (if (and (pair? info) (pair? (cdr info)))
                         (list (car info))
                         (list (x->keyword var))))
                 (default (if (pair? info)
                            (if (pair? (cdr info)) (cadr info) (car info))
                            #f)))
             (list var default keys
                   (cond ((pair? act)
                          (if (not (null? (cddr act)))
                            (errorf "multiple exprs found after =>: ~S" x))
                          `(,(cadr act) (cadr ,l)))
                         (else
                          `(set! ,var (cadr ,l))))))))))
    (else (error "bad binding in let-plist*:" x))))

(define-macro (let-keywords* ls vars . body)
  (let ((undef (gensym))
        (lp (gensym))
        (l (gensym)))
    (if (proper-list? vars)
      (let ((var&default&keys&act (map (lambda (v) (extract-spec v l)) vars)))
        ;; simple loop
        `(let ((,undef (list 'undef)))
           (let (,@(map (lambda (x) `(,(car x) ,undef)) var&default&keys&act))
             (let ,lp ((,l ,ls))
               (cond
                 ((not (null? ,l))
                  (case (car ,l)
                    ,@(map cddr var&default&keys&act))
                  (,lp (cddr ,l)))
                 (else
                  ,@(map (lambda (x)
                           `(if (eq? ,(car x) ,undef)
                              (set! ,(car x) ,(cadr x))))
                         var&default&keys&act)
                  ,@body))))))
      (with-split-improper-list vars
        (lambda (vars2 others-var)
          (let ((var&default&keys&act (map (lambda (v) (extract-spec v l)) vars2)))
            ;; lp accumulating unknown keywords
            `(let ((,undef (list 'undef)))
               (let (,@(map (lambda (x) `(,(car x) ,undef)) var&default&keys&act))
                 (let ,lp ((,l ,ls) (acc '()))
                   (cond
                     ((null? ,l)
                      ,@(map (lambda (x)
                               `(if (eq? ,(car x) ,undef)
                                  (set! ,(car x) ,(cadr x))))
                             var&default&keys&act)
                      (let ((,others-var (reverse! acc)))
                        ,@body))
                     (else
                      (,lp (cddr ,l)
                          (case (car ,l)
                            ,@(map (lambda (x) (append (cddr x) '(acc)))
                                   var&default&keys&act)
                            (else
                             (cons (cadr ,l) (cons (car ,l) acc))))))))))))))))
