
(require-extension (srfi 10000))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(common-module ()
 ((import-extension (common (env file io pathname process tcp)))
  ;; need to standardize an exception system as (common exceptions)
  (cond (sisc (import-extension (srfi 18))))
  (cond (guile
         (prelude
          (define (try thunk . o)
            (catch #t thunk (lambda (e) (if (pair? o) ((car o) e) #f))))))
        (gauche
         (prelude
          (define (try thunk . o)
            (with-error-handler (lambda (e) (if (pair? o) ((car o) e) #f))
              thunk))))
        (else
         (prelude
          (define (try thunk . o)
            (call-with-current-continuation
             (lambda (return)
               (let ((prev-handler (current-exception-handler)))
                 (with-exception-handler
                     (lambda (e)
                       (with-exception-handler prev-handler
                         (lambda () (return (if (pair? o) ((car o) e) #f)))))
                   thunk))))))))
  (entry-point main))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; settings

(define CC "gcc")
(define CSI "csi")
(define CSC "csc")
(define CSC-OPTS '(-O2))
(define CHICKEN-SETUP "chicken-setup")
(define CHICKEN-MODULES-DIR #f)
(define GAMBIT-DIR "/usr/local/Gambit-C")
(define GSI (string-append GAMBIT-DIR "/bin/gsi-script"))
(define GSC (string-append GAMBIT-DIR "/bin/gsc"))
(define GSC-OPTS '())
(define GAMBIT-PRELUDE "(include \"~~/srfi-10000.scm\")")
(define GAMBIT-MODULES-DIR #f)
(define GOSH "gosh")
(define GAUCHE-CONFIG "gauche-config")
(define GAUCHE-MODULES-DIR #f)
(define GUILE "guile")
(define GUILE-CONFIG "guile-config")
(define GUILE-MODULES-DIR #f)
(define SISC "sisc")
(define SISC-MODULES-DIR #f)

(define BIN-DIR "/usr/local/bin")

(define verbose? #f)
(define silent? #f)
(define recursive? #f)
(define force-build? #f)
(define no-action? #f)
(define keep-intermediate-files? #f)
(define default-implementation #f)
(define common-directory #f)
(define repository-cache #f)
(define build-directory #f)
(define repositories '("http://synthcode.com/cs/"))
(define hidden-fields '(export export-syntax entry-point keywords))
(define import-fields
  '(import import-syntax import-extension import-lazy
    import-prefix import-rename inherit))

(define *supported-implementations*
  '(chicken gambit gauche guile sisc))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (main args)
  (if (or (null? args) (null? (cdr args)))
    (command/help)
    (let ((com (string->symbol (cadr args))))
      (load-config com)
      (let lp ((ls (cddr args)) (impl #f))
        (if (or (null? ls)
                (equal? "" (car ls))
                (equal? "--" (car ls))
                (not (eqv? #\- (string-ref (car ls) 0))))
          (let ((impl (or impl default-implementation
                          (current-implementation))))
            (if (not (supported-implementation? impl))
              (die 3 "Unsupported implementation: " impl)
              ((case com
                 ((help) command/help)
                 ((build) command/build)
                 ((install) command/install)
                 ((script) command/script)
                 ((update) command/update)
                 ((upgrade) command/upgrade)
                 ((remove) command/remove)
                 ((search) command/search)
                 ((show) command/show)
                 ((index) command/index)
                 ((list) command/list)
                 (else
                  (die 1 "Unknown command: " (cadr args))))
               impl (if (and (pair? ls) (equal? "--" (car ls))) (cdr ls) ls))))
          (let ((opt
                 (string->symbol
                  (substring (car ls)
                             (if (and (not (equal? "-" (car ls)))
                                      (eqv? #\- (string-ref (car ls) 1)))
                               2 1)
                             (string-length (car ls))))))
            (case opt
              ((all chicken gambit gauche guile sisc)
               (lp (cdr ls) opt))
              ((v verbose) (set! verbose? #t) (lp (cdr ls) impl))
              ((s silent) (set! silent? #t) (lp (cdr ls) impl))
              ((r recursive) (set! recursive? #t) (lp (cdr ls) impl))
              ((n no-action) (set! no-action? #t) (lp (cdr ls) impl))
              ((f force-build) (set! force-build? #t) (lp (cdr ls) impl))
              ((k keep keep-files keep-intermediate-files)
               (set! keep-intermediate-files? #t) (lp (cdr ls) impl))
              (else
               (die 2 "Unknown option: " (car args))))))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Config files - single alist per file, multiple files are appended
;; together.

(define (read-from-file file . o)
  (try (lambda () (with-input-from-file file read))))

(define (load-config-file file)
  (let ((res (read-from-file file '())))  
    (cond
      ((or (pair? res) (null? res))
       (chatter "loaded config file: " file)
       res)
      (else
       (chatter "invalid config file: " file)
       '()))))

(define (assq-ref-atom ls key . o)
  (let ((x (assq key ls)))
    (if x
      (if (pair? (cdr x)) (cadr x) (cdr x))
      (and (pair? o) (car o)))))

(define (assq-ref-list ls key . o)
  (let ((x (assq key ls)))
    (if x
      (if (pair? (cdr x))
        (if (and (pair? (cadr x)) (null? (cddr x))) (cadr x) (cdr x))
        (list (cdr x)))
      (if (pair? o) (car o) '()))))

(define (assoc* key ls . o)
  (let ((eq (if (pair? o) (car o) equal?)))
    (let lp ((ls ls))
      (and (pair? ls)
           (if (eq key (caar ls))
             (car ls)
             (lp (cdr ls)))))))

(define (assoc-ref-atom ls key . o)
  (let-optionals* o ((default #f) (eq equal?))
    (let ((x (assoc* key ls eq)))
      (if x (if (pair? (cdr x)) (cadr x) (cdr x)) default))))

(define (assoc-ref-list ls key . o)
  (let-optionals* o ((default '()) (eq equal?))
    (let ((x (assoc* key ls eq)))
      (if x
        (if (pair? (cdr x))
          (if (and (pair? (cadr x)) (null? (cddr x))) (cadr x) (cdr x))
          (list (cdr x)))
        default))))

(define (load-config command)
  (let lp ((ls (append (load-config-file
                        (path-expand "~/.common-scheme/config"))
                       (load-config-file "/etc/common-scheme/config"))))
    (define (ref-list)
      (if (or (pair? (cdar ls)) (null? (cdar ls)))
        (cdar ls)
        (list (cdar ls))))
    (define (ref-atom) (if (pair? (cdar ls)) (cadar ls) (cdar ls)))
    (define (ref-bool) (and (ref-atom) #t))
    (if (pair? ls)
      (begin
        (if (pair? (car ls))
          (case (caar ls)
            ((verbose?) (set! verbose? (ref-bool)))
            ((silent?) (set! silent? (ref-bool)))
            ((recursive?) (set! recursive? (ref-bool)))
            ((force-build?) (set! force-build? (ref-bool)))
            ((no-action?) (set! no-action? (ref-bool)))
            ((keep-intermediate-files?)
             (set! keep-intermediate-files? (ref-bool)))
            ((default-implementation)
             (set! default-implementation (ref-atom)))
            ((repositories) (set! repositories (ref-list)))
            ((hidden-fields) (set! hidden-fields (ref-list)))
            ((c-compiler) (set! CC (ref-atom)))
            ((chicken-compiler) (set! CSC (ref-atom)))
            ((chicken-compiler-options) (set! CSC-OPTS (ref-list)))
            ((chicken-interpreter) (set! CSI (ref-atom)))
            ((chicken-setup) (set! CHICKEN-SETUP (ref-atom)))
            ((chicken-modules-dir) (set! CHICKEN-MODULES-DIR (ref-atom)))
            ((gambit-compiler) (set! GSC (ref-atom)))
            ((gambit-compiler-options) (set! GSC (ref-list)))
            ((gambit-interpreter) (set! GSI (ref-atom)))
            ((gambit-modules-dir) (set! GAMBIT-MODULES-DIR (ref-atom)))
            ((gosh gauche gauche-interpreter) (set! GOSH (ref-atom)))
            ((gauche-config) (set! GAUCHE-CONFIG (ref-atom)))
            ((gauche-modules-dir) (set! GAUCHE-MODULES-DIR (ref-atom)))
            ((guile guile-interpreter) (set! GUILE (ref-atom)))
            ((guile-config) (set! GUILE-CONFIG (ref-atom)))
            ((guile-modules-dir) (set! GUILE-MODULES-DIR (ref-atom)))
            ((sisc) (set! SISC (ref-atom)))
            ((sisc-modules-dir) (set! SISC-MODULES-DIR (ref-atom)))
            ((bin bin-dir) (set! BIN-DIR (ref-atom)))
            (else
             (if (eq? (caar ls) command)
               (lp (cdar ls))))))
        (lp (cdr ls))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; configuration dependent utilities

(define (chatter . args)
  (if verbose? (apply message args)))

(define (message . args)
  (if (not silent?)
    (with-output-to-port (current-error-port)
      (lambda () (apply pr args) (newline)))))

(define (die n . args)
  (with-output-to-port (current-error-port)
    (lambda () (pr args) (newline))
    (if (< n 2) (command/help))
    (exit n)))

(define (delete-intermediate-file file)
  (if (and (not keep-intermediate-files?) (file-exists? file))
    (delete-file file)))

(define (file-newer? a b)
  (or force-build?
      (or (not (file-exists? a)) (not (file-exists? b)))
      (>= (file-modification-time a) (file-modification-time b))))

(define (exec . o)
  (chatter (string-intersperse (map shell-escape (map display-to-string o))))
  (apply process-run-and-exit o))

(define (shell . o)
  (chatter (string-intersperse (map shell-escape (map display-to-string o))))
  (apply process-run o))

(define (repository-cache-read-location)
  (or repository-cache
      (if common-directory
        (path-build common-directory "index")
        (let ((local (path-expand "~/.common-scheme/index")))
          (if (file-exists? local)
            local
            (path-expand "~root/.common-scheme/index"))))))

(define (repository-cache-write-location)
  (or repository-cache
      (path-build (or common-directory (path-expand "~/.common-scheme"))
                  "index")))

(define (read-repository-cache)
  (read-from-file (repository-cache-read-location) '()))

(define (build-directory-location str)
  (path-build (or build-directory
                  (path-build (or common-directory
                                  (path-expand "~/.common-scheme"))
                              "build"))
              (path-strip-directory str)))

(define (supported-implementation? impl)
  (or (eq? impl 'all) (memq impl *supported-implementations*)))

(define (installed-implementations)
  (let ((imd implementation-modules-directory))
    (append
     (if (file-exists? (path-build (imd 'chicken) "srfi-10000.scm"))
       '(chicken)
       '())
     (if (file-exists? (path-build GAMBIT-DIR "srfi-10000.scm"))
       '(gambit)
       '())
     (if (file-exists? (path-build (imd 'gauche) "srfi-10000.scm"))
       '(gauche)
       '())
     (if (file-exists? (path-build (imd 'guile) "srfi-10000.scm"))
       '(guile)
       '())
     (if (find-in-path
          "sisc"
          (string-split-char (or (getenv "PATH") "/usr/bin") #\:))
       '(sisc)
       '())
     )))

(define (implementation-library-extension impl)
  (case impl
    ((chicken) "so")
    ((gambit) "o1")
    ((sisc) "scc")
    (else "scm")))

(define (implementation-modules-directory impl)
  (case impl
    ((chicken)
     (or CHICKEN-MODULES-DIR
         (process-output->string CHICKEN-SETUP '-repository)
         "/usr/local/lib/chicken"))
    ((gauche)
     (or GAUCHE-MODULES-DIR
         (process-output->string GAUCHE-CONFIG '--sitelibdir)
         "/usr/local/share/gauche/site/lib"))
    ((gambit)
     (or GAMBIT-MODULES-DIR (path-build GAMBIT-DIR "modules")))
    ((guile)
     (or GUILE-MODULES-DIR
         (let ((d (process-output->string GUILE-CONFIG 'info 'pkgdatadir)))
           (if (not (equal? "" d))
             (path-build d "site")
             "/usr/local/share/guile/site"))))
    ((sisc)
     (or SISC-MODULES-DIR "/usr/local/lib/sisc"))
    ))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; string and io utilities

(define (display-to-string x)
  (cond ((string? x) x)
        ((symbol? x) (symbol->string x))
        ((number? x) (number->string x))
        (else (with-output-to-string (lambda () (display x))))))

(define (write-to-string x)
  (with-output-to-string (lambda () (write x))))

(define (string-scan c str . o)
  (let ((pred (if (char? c) (lambda (x) (eqv? x c)) c)))
    (let ((limit (string-length str)))
      (let lp ((i (or (and (pair? o) (car o)) 0)))
        (cond ((>= i limit) #f)
              ((pred (string-ref str i)) i)
              (else (lp (+ i 1))))))))

(define (string-scan-right c str . o)
  (let ((pred (if (char? c) (lambda (x) (eqv? x c)) c)))
    (let lp ((i (or (and (pair? o) (car o)) (- (string-length str) 1))))
      (cond ((negative? i) #f)
            ((pred (string-ref str i)) i)
            (else (lp (- i 1)))))))

(define (string-intersperse ls . o)
  (let ((sep (if (pair? o) (car o) " ")))
    (if (null? ls)
      ""
      (let lp ((ls (cdr ls)) (res (list (car ls))))
        (if (null? ls)
          (apply string-append (reverse res))
          (lp (cdr ls) (cons (car ls) (cons sep res))))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; module utilities

(define (any1 pred ls)
  (let lp ((ls ls)) (and (pair? ls) (or (pred (car ls)) (lp (cdr ls))))))

(define (find1 pred ls)
  (let lp ((ls ls))
    (and (pair? ls) (if (pred (car ls)) ls (lp (cdr ls))))))

(define (remove1 pred ls)
  (let lp ((ls ls) (res '()))
    (if (null? ls)
      (reverse res)
      (lp (cdr ls) (if (pred (car ls)) res (cons (car ls) res))))))

(define (append-map1 f ls)
  (let lp ((ls ls) (res '()))
    (if (null? ls)
      (reverse res)
      (lp (cdr ls) (append (reverse (f (car ls))) res)))))

(define (file-module-declaration file)
  (with-input-from-file file
    (lambda ()
      (if (eqv? #\# (peek-char)) ; just in case they use #!
        (begin (read-char) (if (eqv? #\! (peek-char)) (read-line))))
      (let lp ()
        (let ((x (try read)))
          (and x
               (not (eof-object? x))
               (or (and (pair? x)
                        (eq? 'common-module (car x))
                        (cons (cadr x) (caddr x)))
                   (lp))))))))

(define (library-file? file)
  (let ((ls (file-module-declaration file)))
    (or (not ls)
        (not (find1 (lambda (y)
                      (and (pair? y)
                           (eq? 'entry-point (car y))))
                    (cdr ls))))))

(define (module-name->path name)
  (and (pair? name)
       (let lp ((ls (cdr name)) (res (list (car name))))
         (if (null? ls)
           (apply string-append (map display-to-string (reverse res)))
           (lp (cdr ls) (cons (car ls) (cons "/" res)))))))

(define (module-name->installed-path impl name)
  (path-build (implementation-modules-directory impl)
              (string-append (module-name->path name) ".scm")))

(define (string-split-char str . o)
  (let ((sep (if (pair? o) (car o) #\space))
        (len (string-length str)))
    (let lp ((i 0) (j 0) (res '()))
      (define (collect) (cons (substring str j i) res))
      (cond ((= i len) (reverse (collect)))
            ((eqv? sep (string-ref str i)) (lp (+ i 1) (+ i 1) (collect)))
            (else (lp (+ i 1) j res))))))

(define (module-dot-string->name str)
  (map string->symbol (string-split-char str #\.)))

(define (module-name->dot-string ls)
  (string-intersperse (map display-to-string ls) "."))

(define (module-path->name str)
  (map string->symbol (string-split-char (path-strip-extension str) #\/)))

(define (expand-module-specifier spec)
  (cond
    ((null? spec) (list spec))
    ((not (pair? spec)) (list (list spec)))
    (else
     (append-map1
      (lambda (head)
        (map (lambda (tail) (append head tail))
             (expand-module-specifier (cdr spec))))
      (if (pair? (car spec))
        (append-map1 expand-module-specifier (car spec))
        (list (list (car spec))))))))

(define (module-dependencies x)
  (let ((info (if (string? x) (file-module-declaration x) x)))
    (if (not (pair? info))
      '()
      (let lp ((ls info) (res '()))
        (cond ((not (pair? ls)) res)
              ((pair? (car ls))
               (case (caar ls)
                 ((import-extension import-syntax inherit)
                  (lp (cdr ls)
                      (append (append-map1 expand-module-specifier (cdar ls))
                              res)))
                 ((import-lazy import-prefix import-rename)
                  (lp (cdr ls) (cons (cadar ls) res)))
                 (else (lp (cdr ls) res))))
              (else (lp (cdr ls) res)))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; URL utilities

(define (http-url? str)
  (and (> (string-length str) 5)
       (string-ci=? (substring str 0 5) "http:")))

(define (call-with-url-positions url proc)
  (cond
    ((string-scan #\: url)
     => (lambda (i)
          (cond
            ((string-scan (lambda (c) (not (eqv? c #\/))) url (+ i 1))
             => (lambda (j)
                  (let ((k (or (string-scan #\/ url (+ j 1))
                               (string-length url))))
                    (proc i j k))))
            (else (proc i #f #f)))))
    (else (proc #f #f #f))))

(define (url-schema url)
  (call-with-url-positions url
    (lambda (colon host-start path-start)
      (substring url 0 colon))))

(define (url-host url)
  (call-with-url-positions url
    (lambda (colon host-start path-start)
      (substring url host-start path-start))))

(define (url-path url)
  (call-with-url-positions url
    (lambda (colon host-start path-start)
      (substring url path-start (string-length url)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Very simple http fetch utility, just writes the body of a URL to the
;; optional second argument, defaulting to current-output-port.  If the
;; destination is a string it's used as a file-name to write to.

(define (http-fetch url . o)
  (define (char-not-whitespace? c) (not (char-whitespace? c)))
  (define (read-header)
    (let lp ((enc "none"))
      (let ((line (read-line)))
        (if (or (eof-object? line) (string=? line "") (string=? line "\r"))
          enc
          (let ((i (string-scan #\: line)))
            (if (and i (string-ci=? "transfer-encoding" (substring line 0 i)))
              (lp (substring
                   line
                   (string-scan char-not-whitespace? line (+ i 1))
                   (string-length line)))
              (lp enc)))))))
  (define (read-location)
    (let ((line (read-line)))
      (and (not (eof-object? line))
           (not (string=? line ""))
           (cond
             ((string-scan #\: line)
              => (lambda (i)
                   (if (string-ci=? "location" (substring line 0 i))
                     (substring
                      line
                      (string-scan char-not-whitespace? line (+ i 1))
                      (string-length line))
                     (read-location))))
             (else (read-location))))))
  (define (read-chunked)
    (let lp ()
      (let ((line (read-line)))
        (if (string? line)
          (let ((size (string->number line 16)))
            (if size
              (begin
                (display (read-string-n size))
                (lp))))))))
  (define (read-string-n size) ; XXXX SLOW!!!
    (let lp1 ((i 0) (res '()))
      (if (= i size)
        (apply string (reverse res))
        (let ((c (read-char)))
          (if (eof-object? c)
            (apply string (reverse res))
            (lp (+ i 1) (cons c res)))))))
  (define (read-plain)
    (let lp ()
      (let ((line (read-line)))
        (if (not (eof-object? line))
          (begin
            (display line) (newline)
            (lp))))))
  (define (read-body)
    (if (string-ci=? "chunked" (read-header))
      (read-chunked)
      (read-plain)))
  (let ((orig-out (current-output-port)))
    (with-tcp-io (url-host url) 80
      (lambda ()
        (pr "GET " url " HTTP/1.0\r\n\r\n")
        (flush-output)
        (let ((resp (read-line)))
          (and (string? resp)
               (let* ((i (string-scan #\space resp))
                      (j (and i (string-scan #\space resp (+ i 1))))
                      (n (and j (string->number (substring resp (+ i 1) j)))))
                 (if n
                   (case n
                     ((200)
                      (let* ((p0 (if (pair? o) (car o) orig-out))
                             (p (if (string? p0) (open-output-file p0) p0)))
                        (with-output-to-port p read-body)
                        (if (string? p0) (close-output-port p))))
                     ((302) (apply http-fetch (read-location) o))
                     (else (error "couldn't download url:" url)))))))))))

(define (http-fetch-file-to-build url)
  (let* ((file (path-strip-directory url))
         (dest (build-directory-location
                (if (equal? file "") "index.html" file))))
    (create-directory* (path-directory dest))
    (call-with-output-file dest (lambda (p) (http-fetch url p)))
    dest))

;; returns (repo . module-info)
(define (find-module-in-cache name . o)
  (let lp ((ls (if (pair? o) (car o) (read-repository-cache))))
    (and (pair? ls)
         (let* ((mods (assq-ref-list (cdar ls) 'modules))
                (info (assoc-ref-list mods name)))
           (if (pair? info)
             (cons (caar ls) info)
             (lp (cdr ls)))))))

;; return a local path for an arbitrary module string, downloading if
;; needed
(define (module-local-path str)
  (cond
    ((http-url? str) (http-fetch-file-to-build str))
    ((file-exists? str) str)
    (else
     (let* ((name (module-dot-string->name str))
            (path (string-append (module-name->path name) ".scm")))
       (if (file-exists? path)
         path
         (let ((info (find-module-in-cache name)))
           (if info
             (let* ((remote-path (assoc-ref-atom (cdr info) 'path path))
                    (repo (car info))
                    (url
                     (cond
                       ((http-url? remote-path)
                        remote-path)
                       ((eqv? #\/ (string-ref repo (- (string-length repo) 1)))
                        (path-expand remote-path repo))
                       (else
                        (path-expand remote-path (path-directory repo))))))
               (http-fetch-file-to-build url))
             (die 5 "Can't find module: " str))))))))

(define (expand-to-local-files impl ls)
  (if (not recursive?)
    (map module-local-path ls)
    (let ((installed (installed-modules impl))
          (available (append-map1
                      (lambda (repo)
                        (let ((ls (assq-ref-list (cdr repo) 'modules)))
                          (if ls (map car ls) '())))
                      (read-repository-cache))))
      (let lp ((ls ls) (res '()) (mods '()))
        (if (null? ls)
          res
          (let ((str (car ls)))
            (cond
              ((http-url? str)
               (lp (cdr ls) (cons (http-fetch-file-to-build str) res) mods))
              ((file-exists? str)
               (let* ((info (file-module-declaration str))
                      (name (and (pair? info) (car info)))
                      (rel-dir
                       (and (or (pair? name) (null? name))
                            (let lp ((i (length name)) (res str))
                              (if (negative? i)
                                res
                                (lp (- i 1) (path-directory res))))))
                      (deps (if (pair? info) (module-dependencies info) '())))
                 (let lp2 ((deps deps) (new-deps '()) (new-mods '()))
                   (if (null? deps)
                     (lp (append new-deps (cdr ls))
                         (cons str res)
                         (append new-mods mods))
                     (let ((name2 (car deps)))
                       (if (member name2 installed)
                         (lp2 (cdr deps) new-deps new-mods)
                         (let ((path (and rel-dir
                                          (path-build
                                           rel-dir
                                           (string-append
                                            (module-name->path name2)
                                            ".scm")))))
                           (if (and (not (member path ls))
                                    (not (member path res))
                                    (file-exists? path))
                             (lp2 (cdr deps) (cons path new-deps) new-mods)
                             (let ((dot-name (module-name->dot-string name2)))
                               (if (and (not (member dot-name ls))
                                        (not (member dot-name mods))
                                        (member dot-name available))
                                 (lp2 (cdr deps)
                                      (cons dot-name new-deps)
                                      (cons dot-name new-mods))
                                 (lp2 (cdr deps) new-deps new-mods)))))))))))
              (else
               (let ((name (module-dot-string->name str)))
                 (if (member name installed)
                   (begin
                     (chatter "module " str " already installed")
                     (lp (cdr ls) res mods))
                   (let ((file (module-local-path str)))
                     (lp (if file (cons file (cdr ls)) (cdr ls))
                         res
                         mods))))))))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; commands

(define (command/script impl ls)
  (cond
    ((null? ls)
     (command/help))
    ((eq? impl 'all)
     (for-each (lambda (impl) (script impl ls)) (installed-implementations)))
    (else
     (chatter "script [" impl "] " ls)
     (let ((f (car (expand-to-local-files impl (list (car ls))))))
       (case impl
         ((chicken) (apply exec CSI '-script f (cdr ls)))
         ((gambit)  (apply exec GSI '-e GAMBIT-PRELUDE f (cdr ls)))
         ((gauche)  (apply exec GOSH f (cdr ls)))
         ((guile)   (apply exec GUILE '-s f (cdr ls)))
         ((sisc)    (apply exec SISC '-c 'common-main '-x f '-- (cdr ls)))
         )))))

(define (gambit-build file)
  (let* ((file (path-expand file))
         (base (path-strip-extension file))
         (local (path-strip-directory base)))
    (cond
      ((library-file? file)
       (let ((dest (string-append base ".o1")))
         (cond
           ((file-newer? file dest)
            (shell GSC '-dynamic '-prelude GAMBIT-PRELUDE file)
            (delete-intermediate-file (string-append local ".c"))
            (rename-file (string-append local ".o1") dest)))))
      (else
       (let ((dest base))
         (cond
           ((file-newer? file dest)
            ;; XXXX Gambit currently can't compile outside pwd
            (shell (string-append "cd " (path-directory file) "; "
                                  (display-to-string GSC))
                   '-prelude GAMBIT-PRELUDE (path-strip-directory file))
            (shell CC '-O2 '-I/usr/local/Gambit-C/include
                   '-L/usr/local/Gambit-C/lib
                   (string-append base ".c") (string-append base "_.c")
                   '-lgambc '-lm '-ldl '-lutil '-o dest)
            (delete-intermediate-file (string-append base ".c"))
            (delete-intermediate-file (string-append base "_.c"))
            )))))))

(define (command/build impl files)
  (let ((files (expand-to-local-files impl files)))
    (cond
      ((eq? impl 'all)
       ;; For now it's safe to build all at once since the target
       ;; extensions are different.
       (for-each
        (lambda (impl) (command/build impl files))
        (installed-implementations)))
      (else
       (message "build [" impl "] " files)
       (case impl
         ((chicken)
          (for-each
           (lambda (f)
             (if (library-file? f)
               (let ((dest (path-replace-extension f "so")))
                 (if (file-newer? f dest)
                   (apply shell CSC '-s (append CSC-OPTS (list f)))))
               (let ((dest (path-strip-extension f)))
                 (if (file-newer? f dest)
                   (apply shell CSC (append CSC-OPTS (list f)))))))
           files))
         ((gambit)
          (for-each gambit-build files))
         ((sisc)
          (let lp ((ls files) (res '()))
            (if (null? ls)
              (let ((expr (cons 'begin
                                (cons `(require-extension (srfi 10000))
                                      (reverse res)))))
                (shell SISC '-x '-e (write-to-string expr)))
              (let* ((src (car ls))
                     (dest (path-replace-extension src "scc")))
                (if (file-newer? src dest)
                  (lp (cdr ls) (cons `(compile-file ,src ,dest) res))
                  (lp (cdr ls) res))))))
         )))))

(define (command/install impl files)
  (let ((files (expand-to-local-files impl files)))
    (set! recursive? #f)
    (command/build impl files)  ; install implies build
    (cond
      ((eq? impl 'all)
       (for-each
        (lambda (impl) (command/install impl files))
        (installed-implementations)))
      (else
       (message "install [" impl "] " files)
       (case impl
         ((chicken gambit sisc)
          (for-each
           (lambda (file)
             (cond
               ((library-file? file)
                (let* ((ext (implementation-library-extension impl))
                       (src (path-replace-extension file ext))
                       (ls (file-module-declaration file))
                       (name (car ls))
                       (info (cdr ls))
                       (dest (string-append (module-name->path name) "." ext))
                       (mod-dir (implementation-modules-directory impl))
                       (abs-dest (path-build mod-dir dest))
                       (dest-dir (path-directory abs-dest))
                       (setup (path-replace-extension abs-dest "setup"))
                       (syntax? (assq 'export-syntax info)))
                  (create-directory* dest-dir)
                  (rename-file src abs-dest)
                  (if (or syntax? (eq? impl 'chicken))
                    (copy-file file dest-dir))
                  (with-output-to-file setup
                    (lambda ()
                      (write `((files ,abs-dest)
                               (module)
                               ,@(if syntax? '((syntax)) '())
                               ,@info))))))
               (else
                (let* ((prog (path-strip-extension file))
                       (bin 
                        (case impl
                          ((chicken)
                           (process-output->string
                            CHICKEN-SETUP 'program-path))
                          (else BIN-DIR)))
                       (dest (path-build bin (path-strip-directory prog))))
                  (rename-file prog dest)))))
           files))
         ((gauche guile)
          (for-each
           (lambda (file)
             (cond
               ((library-file? file)
                (let* ((ls (file-module-declaration file))
                       (name (car ls))
                       (dest (string-append (module-name->path name) ".scm"))
                       (mod-dir (implementation-modules-directory impl))
                       (abs-dest (path-build mod-dir dest))
                       (dest-dir (path-directory abs-dest)))
                  (create-directory* dest-dir)
                  (copy-file file abs-dest)))
               (else
                (let* ((prog (path-strip-extension file))
                       (dest (path-build BIN-DIR (path-strip-directory prog))))
                  (copy-file file dest)))))
           files))
         )))))

(define (command/remove impl ls)
  (if (eq? impl 'all)
    (for-each (lambda (i) (command/remove i ls)) (installed-implementations))
    (for-each
     (lambda (mod)
       (let ((f (path-replace-extension
                 (module-name->installed-path
                  impl (module-dot-string->name mod))
                 (implementation-library-extension impl))))
         (message "remove [" impl "] " f)
         (if (and (not no-action?) (file-exists? f))
           (begin
             (delete-file f)
             (let ((setup (path-replace-extension f "setup")))
               (if (file-exists? setup)
                 (delete-file setup)))))))
     ls)))

(define (command/update impl ls)
  (define (fetch-repository str)
    (message ";updating repository: " str)
    (try
     (lambda ()
       (let* ((x (with-output-to-string (lambda () (http-fetch str))))
              (repo (with-input-from-string x read)))
         (cons str (if (pair? repo) repo '()))))
     (lambda (e) (message ";failed to fetch repository: " str) #f)))
  (define (fetch-all-repositories ls)
    (if recursive?
      (let lp ((ls ls) (res '()))
        (if (null? ls)
          res
          (let ((repo (fetch-repository (car ls))))
            (if (pair? repo)
              (let ((peers (append-map1
                            (lambda (x)
                              (if (and (pair? x) (eq? 'peers (car x)))
                                (if (pair? (cdr x)) (cdr x) (list (cdr x)))
                                '()))
                            (cdr repo)))
                    (res (cons repo res)))
                (lp (append
                     (remove1 (lambda (x) (or (assoc x res) (member x ls)))
                              peers)
                     (cdr ls))
                    res))
              (lp (cdr ls) res)))))
      (remove1 not (map fetch-repository ls))))
  (let* ((file (repository-cache-write-location))
         (d (message "file: " file))
         (cache (read-from-file file '()))
         (cache (if (pair? cache) cache '())))
    (message "read cache")
    (create-directory* (path-directory file))
    (with-output-to-file file
      (lambda ()
        (if (null? ls)
          (if (null? cache) ; handle first-time
            (write (fetch-all-repositories repositories))
            (write (fetch-all-repositories (map car cache))))
          (let ((updates (fetch-all-repositories ls)))
            (write (append
                    (map (lambda (x) (or (assoc (car x) updates) x)) cache)
                    (remove1 (lambda (x) (assoc (car x) cache)) updates)))))))))

(define (installed-modules impl)
  (let ((ext (implementation-library-extension impl))
        (dir (implementation-modules-directory impl)))
    (map
     (lambda (path)
       (let ((rel (path-relative path dir)))
         (if rel (module-path->name rel) '())))
     (directory-fold-tree
      dir
      (lambda (d f a)
        (let ((res
               (try
                (lambda ()
                  (and (equal? ext (path-extension f))
                       (or (not (equal? ext "scm"))
                           (pair? (file-module-declaration
                                   (path-build d f))))
                       (path-build d f))))))
          (if res (cons res a) a)))
      '()))))

(define (command/upgrade impl ls)
  (if (eq? impl 'all)
    (for-each
     (lambda (i) (command/upgrade i ls))
     (installed-implementations))
    (for-each
     (lambda (m)
       (let ((info (find-module-in-cache m)))
         (if info
           (let* ((latest-time (assq-ref-atom (cdr info) 'modified 0))
                  (path (path-replace-extension
                         (module-name->installed-path impl m)
                         (implementation-library-extension impl)))
                  (installed-time (if (and path (file-exists? path))
                                    (file-modification-time path)
                                    0)))
             (if (> latest-time installed-time)
               (command/install impl (list (string-intersperse
                                            (map display-to-string m)
                                            "."))))))))
     (if (pair? ls)
       (map module-dot-string->name ls)
       (installed-modules impl)))))

(define (command/search impl ls)
  (define (string-contains-ci a b) ; slow
    (let ((a-len (string-length a))
          (b-len (string-length b)))
      (let lp1 ((i 0))
        (and (< i a-len)
             (let lp2 ((i2 i) (j 0))
               (cond ((= j b-len) i)
                     ((char-ci=? (string-ref a i2) (string-ref b j))
                      (lp2 (+ i2 1) (+ j 1)))
                     (else (lp1 (+ i 1)))))))))
  (let lp ((cache (read-repository-cache)))
    (if (pair? cache)
      (begin
        (for-each
         (lambda (mod)
           (let* ((name (car mod))
                  (name-string
                   (string-intersperse (map display-to-string name) "."))
                  (desc (assq-ref-atom (cdr mod) 'description))
                  (summary (assq-ref-atom (cdr mod) 'summary)))
             (if (or (null? ls)
                     (any1 (lambda (s)
                             (or (string-contains-ci name-string s)
                                 (and desc (string-contains-ci desc s))
                                 (and summary (string-contains-ci summary s))))
                           ls))
               (pr name-string
                   " - "
                   (or summary
                       (and desc
                            (with-input-from-string desc read-line))
                       "")
                   nl))))
         (assq-ref-list (cdar cache) 'modules))
        (lp (cdr cache))))))

(define (command/show impl mods)
  (define (title-case x)
    (with-string-io (display-to-string x)
      (lambda ()
        (let lp ((boundary? #t))
          (let ((c (read-char)))
            (if (not (eof-object? c))
              (if (or (char-alphabetic? c) (char-numeric? c))
                (let ((c (if boundary? (char-upcase c) (char-downcase c))))
                  (write-char c)
                  (lp #f))
                (begin (write-char c) (lp #t)))))))))
  (define (show x)
    (if (and (pair? x)
             (or verbose? (not (memq (car x) hidden-fields)))
             (not (memq (car x) import-fields)))
      (pr (title-case (car x)) ": "
          (if (and (pair? (cdr x)) (null? (cddr x)))
            (cadr x)
            (cdr x))
          nl)))
  (let ((cache (read-repository-cache)))
    (for-each
     (lambda (name)
       (let ((info (find-module-in-cache name)))
         (if (pair? info)
           (let* ((deps (map module-name->dot-string
                             (module-dependencies (cdr info))))
                  (repo (assoc-ref-list cache (car info)))
                  (defaults (assq 'default-fields repo))
                  (new-defaults
                    (remove1
                     (lambda (x) (or (not (pair? x)) (assoc (car x) (cdr info))))
                     (cdr defaults))))
             (pr "Module: " (module-name->dot-string name) nl
                 "Repository: " (car info) nl)
             (if (pair? deps)
               (pr "Requires: "
                   (string-intersperse (map display-to-string deps) ", ")
                   nl))
             (for-each show (append new-defaults (cdr info)))))))
     (map module-dot-string->name mods))))

(define (command/list impl mods)
  (if (eq? impl 'all)
    (for-each
     (lambda (impl)
       (pr impl ":\n")
       (command/list impl mods))
     (installed-implementations))
    (for-each
     (lambda (x) (pr x nl))
     (map module-name->dot-string (installed-modules impl)))))

(define (command/index impl ls)
  (let-optionals* ls ((dir ".")
                      (file (path-build dir "index.txt")))
    (define (writeln pre x)
      (display pre) (write x) (newline))
    (define (kons d f knil)
      (try
       (lambda ()
         (if (equal? "scm" (path-extension f))
           (let* ((path (path-build d f))
                  (info (file-module-declaration path)))
             (if (and (pair? info) (pair? (car info)))
               (begin
                 (message ";indexing " (car info) ": " path)
                 (pr "  (" (car info) nl)
                 (let ((rel (path-relative path dir)))
                   (if rel (writeln "   " (list 'path rel))))
                 (writeln "   " (list 'modified (file-modification-time path)))
                 (for-each (lambda (x) (writeln "   " x)) (cdr info))
                 (pr "   )\n"))))))
       (lambda (e) (message ";error indexing file: " str) #f)))
    (define (go . o)
      (if (not no-action?)
        (begin
          (display "(\n")
          (if (pair? o)
            (for-each (lambda (x) (writeln " " x)) (car o)))
          (display " (modules\n")))
      (directory-fold-tree dir kons #f)
      (if (not no-action?) (pr "  ))\n")))
    (if no-action?
      (go)
      (if (file-exists? file)
        (let ((ls (read-from-file file)))
          (if (or (not (pair? ls)) (not (assq 'modules ls)))
            (die 4 "File '" file
                 "' doesn't look like an index, not overwriting.")
            (with-output-to-file file
              (lambda ()
                (go (remove1 (lambda (x)
                               (or (not (pair? x)) (eq? 'modules (car x))))
                             ls))))))
        (with-output-to-file file go)))))

(define (command/help . o)
  (display
"usage: common-scheme <command> [options ...] [args ...]

Commands:
  script file [args ...]      run a file as a script, passing args
  build files ...             compile files given on command-line
  install files ...           install files given on command-line
  update [repositories ...]   retrieve new lists of modules
  upgrade [modules ...]       upgrade some or all outdated modules
  remove modules ...          remove installed modules
  search keywords ...         search for modules containing keywords
  show modules ...            show detailed module information
  index [dir [file]]          create a repository index file
  list                        list installed modules
  help                        this message

Options may include at most one implementation:
  --all --chicken --gambit --gauche --guile --sisc
The default is the implementation common-scheme was compiled or run with.

Miscellaneous options:
  -v, --verbose      -  report extra information
  -s, --silent       -  keep output to a minimum
  -r, --recursive    -  recursively build/install dependent files
  -n, --no-action    -  just print what actions would be performed
  -k, --keep-files   -  don't clean up intermediate files
  -f, --force-build  -  build even if the file is up-to-date
"))

)
